"""Utilities for sending log messages to a central ELK."""

import datetime
import logging
import traceback
from logging import LogRecord
from typing import TYPE_CHECKING

import jsons
from enochecker_core import EnoLogMessage

LOGGING_PREFIX = "##ENOLOGMESSAGE "

if TYPE_CHECKING:  # pragma: no cover
    from .enochecker import BaseChecker


def exception_to_string(excp: Exception) -> str:
    """
    Format an exception as a string.

    Limits the length of the traceback to 3.
    :param ecxp: the exception to format
    :return: The formatted string
    """
    stack = traceback.extract_stack()[:-3] + traceback.extract_tb(
        excp.__traceback__
    )  # add limit=??
    pretty = traceback.format_list(stack)
    return "".join(pretty) + f"\n  {excp.__class__} {excp}"


class ELKFormatter(logging.Formatter):
    """Format log messages for a central ELK."""

    def __init__(
        self,
        checker: "BaseChecker",
        fmt: str = None,
        datefmt: str = "%Y-%m-%dT%H:%M:%S%z",
        style: str = "%",
    ) -> None:
        """
        Initialize a new formatter.

        :param checker: The checker instance to which this formatter belongs, needed to include information about the service in the messages.
        :param fmt: format string passed to :class:`logging.Formatter`
        :param datefmt: date format string passed to :class:`logging.Formatter`
        :param style: style string passed to :class:`logging.Formatter`
        """
        super().__init__(fmt, datefmt, style)
        self.checker: "BaseChecker" = checker

    def format(self, record: LogRecord) -> str:
        """
        Format a LogRecord as a string.

        :param record: the record to format
        :return: the formatted string
        """
        record.asctime = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%S.%fZ")

        message = record.getMessage()
        if record.exc_info:
            eno = record.exc_info
            stacktrace = "".join(traceback.format_exception(None, eno[1], eno[2]))
            message += f" excp: {stacktrace}"
        if record.stack_info:
            stack = self.formatStack(record.stack_info)
            message += f" trace: {stack}"

        log_message = EnoLogMessage(
            tool=type(self.checker).__name__,
            type="infrastructure",
            severity=record.levelname,
            severity_level=max(0, record.levelno // 10 - 1),
            timestamp=record.asctime,
            message=message,
            module=record.module,
            function=record.funcName,
            service_name=self.checker.service_name,
            method=self.checker.method.value if self.checker.method else None,
            task_id=self.checker.task_id,
            team_id=self.checker.team_id,
            team_name=self.checker.team_name,
            current_round_id=self.checker.current_round_id,
            related_round_id=self.checker.related_round_id,
            flag=self.checker.flag,
            variant_id=self.checker.variant_id,
            task_chain_id=self.checker.task_chain_id,
        )

        return LOGGING_PREFIX + jsons.dumps(
            log_message,
            use_enum_name=False,
            key_transformer=jsons.KEY_TRANSFORMER_CAMELCASE,
            strict=True,
        )



"""Flask service to run a checker as HTTP service."""

import logging
import os
from typing import TYPE_CHECKING, Callable, Tuple, Type

import jsons
from enochecker_core import (
    CheckerInfoMessage,
    CheckerResultMessage,
    CheckerTaskMessage,
    CheckerTaskResult,
)
from flask import Flask, Response, request

from .logging import exception_to_string

if TYPE_CHECKING:  # pragma: no cover
    from .enochecker import BaseChecker

logging.basicConfig(level=logging.DEBUG)
logger = logging.Logger(__name__)
logger.setLevel(logging.DEBUG)

with open(os.path.join(os.path.dirname(__file__), "post.html")) as f:
    INDEX_PAGE = f.read()


def checker_routes(
    checker_cls: Type["BaseChecker"],
) -> Tuple[
    Callable[[], Response],
    Callable[[], Response],
    Callable[[], CheckerInfoMessage],
    Callable[[], Response],
]:
    """
    Create a flask app for the given checker class.

    :param checker_cls: The checker class to use
    :return: A flask app that can be passed to a uWSGI server or run using .run().
    """

    def index() -> Response:
        """
        Display general info about this service.

        Includes a web interface for manually sending requests to the service.

        :return: Flask resposne containing the HTML page with info about this service
        """
        logger.info("Request on /", mimetype="text/html")

        return Response(INDEX_PAGE)

    def serve_checker() -> Response:
        """
        Serve a single checker request.

        The request needs to be formed according to the spec of CheckerTaskMessage.

        :return: Flask response containing the CheckerResultMessage as JSON
        """
        try:
            logger.info(request.json)
            try:
                task_msg = jsons.loads(
                    request.get_data(),
                    CheckerTaskMessage,
                    strict=True,
                    key_transformer=jsons.KEY_TRANSFORMER_SNAKECASE,
                )
            except jsons.exceptions.UnfulfilledArgumentError as e:
                return Response(e._msg, status=400)

            checker = checker_cls(task_msg)

            checker.logger.info(task_msg)
            res = checker.run()

            result_message = CheckerResultMessage(
                result=res.result, message=res.message
            )

            res_json = jsons.dumps(
                result_message,
                use_enum_name=False,
                key_transformer=jsons.KEY_TRANSFORMER_CAMELCASE,
                strict=True,
            )

            checker.logger.info("{}".format(res_json))

            return Response(res_json, mimetype="application/json")

        except Exception as ex:
            logger.error(
                "Returning Internal Error {}.\nTraceback:\n{}".format(
                    ex, exception_to_string(ex)
                ),
                exc_info=ex,
            )
            result_message = CheckerResultMessage(
                result=CheckerTaskResult.CHECKER_TASK_RESULT_INTERNAL_ERROR,
                message=f"Critical checker error occured\n{exception_to_string(ex)}",
            )
            res_json = jsons.dumps(
                result_message,
                use_enum_name=False,
                key_transformer=jsons.KEY_TRANSFORMER_CAMELCASE,
                strict=True,
            )
            return Response(res_json, mimetype="application/json")

    def service_info() -> CheckerInfoMessage:
        """
        Return technical information about this service.

        Includes the name and supported variants of flags, havoc and noise.

        :return: CheckerInfoMessage filled with the values from the checker
        """
        try:
            service_name: str = getattr(
                checker_cls, "service_name", checker_cls.__name__.split("Checker")[0]
            )

            return CheckerInfoMessage(
                service_name=service_name,
                flag_variants=checker_cls.flag_variants,
                noise_variants=checker_cls.noise_variants,
                havoc_variants=checker_cls.havoc_variants,
            )

        except Exception:
            logger.error(
                """Service info not specified!
add service_name, flag_variants, havoc_variants and noise_variants as static fields to your CHECKER

Example:
class ExampleChecker(BaseChecker):
    flag_variants  = 1
    noise_variants = 1
    havoc_variants = 1
"""
            )
            raise AttributeError("REQUIRED SERVICE INFO FIELDS NOT SPECIFIED!")

    def get_service_info() -> Response:
        """
        Return a Flask response containing the JSON representation of the information from :func:`service_info`.

        :return: Flask response containing the JSON representation of the service info
        """
        res_json = jsons.dumps(
            service_info(),
            use_enum_name=False,
            key_transformer=jsons.KEY_TRANSFORMER_CAMELCASE,
            strict=True,
        )

        return Response(res_json, mimetype="application/json")

    return index, serve_checker, service_info, get_service_info


def init_service(checker: Type["BaseChecker"]) -> Flask:
    """
    Initialize a flask app that can be used for WSGI or listen directly.

    The Engine may communicate with it over socket.

    :param checker: the checker class to use for check requests.
    :return: a flask app with post and get routes set, ready for checking.
    """
    app = Flask(__name__)
    index, checker_route, service_info, get_service_info = checker_routes(checker)

    app.route("/", methods=["GET"])(index)
    app.route("/", methods=["POST"])(checker_route)
    app.route("/service", methods=["GET"])(get_service_info)

    logger.info(service_info())

    return app  # Start service using service.run(host="0.0.0.0")



"""Backend for team_db based on a local filesystem directory."""

import atexit
import json
import logging
import os
import threading
import time
from collections.abc import MutableMapping
from functools import wraps
from pathlib import Path
from typing import Any, Callable, Dict, Iterator, Optional, Set

from .utils import base64ify, debase64ify, ensure_valid_filename, start_daemon

logging.basicConfig(level=logging.DEBUG)
dictlogger = logging.Logger(__name__)
dictlogger.setLevel(logging.DEBUG)

DB_DEFAULT_DIR = os.path.join(
    os.getcwd(), ".data"
)  # Default location db files will be stored in -> usually cwd.
# TODO: Force remove locks after a while?
DB_LOCK_RETRYCOUNT = (
    6  # 2**6 / 10 seconds are 6.4 secs. -> That's how long the db will wait for a log
)
DB_PREFIX = "_store_"  # Prefix all db files will get
DB_EXTENSION = ".json"  # Extension all db files will get
DB_LOCK_EXTENSION = ".lock"  # Extension all lock folders will get
DB_GLOBAL_CACHE_SETTING = True


def makedirs(path: str, exist_ok: bool = True) -> None:
    """
    Create a directory.

    Creates the parent directories if necessary.

    param path: the path to create
    param exist_ok: ignore already existing path and do nothing
    """
    Path(path).mkdir(parents=True, exist_ok=exist_ok)


def _locked(func: Callable[..., Any]) -> Callable[..., Any]:
    """
    Return a wrapped method for StoredDict accesses to ensure locks on a python threading level.

    :param func: StoredDict method to be wrapped
    :return: the wrapped method
    """

    @wraps(func)
    def locked(self: "StoredDict", *args: str, **kwargs: int) -> Any:
        """
        Wrap the function.

        First acquires a lock and then releases it later.

        :param args: positional arguments passed to the wrapped function
        :param kwargs: kwargs to pass to the wrapped function
        :return: the wrapped function
        """
        self.logger.debug("Locking {} db".format(self.name))
        self._local_lock.acquire()
        self.logger.debug("Log db lock for {}".format(self.name))
        try:
            return func(self, *args, **kwargs)
        finally:
            self._local_lock.release()
            self.logger.debug("Released db lock for {}".format(self.name))

    return locked


class StoredDict(MutableMapping):
    """
    A dictionary that is filesystem backed.

    It will write to disk every few seconds and at exit.
    In case python crashes, changes may be gone. :/
    Note: Complex won't be tracked.
    """

    def __init__(
        self,
        name: str = "default",
        base_path: str = DB_DEFAULT_DIR,
        persist_secs: int = 3,
        ignore_locks: bool = False,
        logger: Optional[logging.Logger] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """
        Create a new File System backed Store.

        It quacks like a dict and will persist to filesystem every few seconds if possible. :)

        :param name: name of this store
        :param base_path: the base path
        :param persist_secs: how often to persist dirty elements (0 means: never autostore. Call persist manually)
        :param ignore_locks: We usually write and read lock files before accessing the data.
                This flag seaves them out.
        :param logger: The logger instance to log events to
        """
        self._cache: Dict[str, Any] = {}
        self._dirties: Set[str] = set()
        self._locks: Set[str] = set()
        self._to_delete: Set[str] = set()
        self._persist_thread: Optional[threading.Thread] = None
        self.name = name
        self.path: str = os.path.join(base_path, ensure_valid_filename(name))
        self.persist_secs: int = persist_secs
        self.ignore_locks: bool = ignore_locks

        if logger:
            self.logger = logger
        else:
            self.logger = dictlogger

        self._local_lock = threading.RLock()

        makedirs(self.path, exist_ok=True)
        atexit.register(self._cleanup)
        self._stopping = False

        self.update(
            dict(*args, **kwargs)
        )  # In case we got initialized using a dict, make sure it's in sync.

    @_locked
    def _spawn_persist_thread(self) -> None:
        """Spawn a thread persisting all changes, if necessary."""
        if self.persist_secs > 0 and not self._persist_thread:

            def persist_async() -> None:
                time.sleep(self.persist_secs)
                self.logger.debug("Persisting db {} from background.".format(self.name))
                self.persist()

            self._persist_thread = start_daemon(persist_async)

    @_locked
    def _cleanup(self) -> None:
        """Clean up the db: persists and releases all locks currently held."""
        self.logger.debug("StoredDict cleanup task running.")
        self._stopping = True
        self.persist()
        for lock in self._locks:
            self.release(lock)

    def __del__(self) -> None:
        """Delete a key."""
        self._cleanup()

    def _dir(self, key: str) -> str:
        """
        Return the path where data for this key is stored.

        :param key: the key to look up
        :return: string representation of the file path
        """
        return os.path.join(self.path, DB_PREFIX + base64ify(key, b"+-"))

    def _dir_jsonname(self, key: str) -> str:
        """
        Return the path for the json db file for this key.

        See :func:`_dir`
        """
        return "{}{}".format(self._dir(key), DB_EXTENSION)

    def _dir_lockname(self, key: str) -> str:
        """
        Return the path for the lock file for this key.

        See :func:`_dir`
        """
        return "{}{}".format(self._dir(key), DB_LOCK_EXTENSION)

    @_locked
    def release(self, locked_key: str) -> None:
        """
        Release a file lock.

        :param locked_key: the key we locked
        """
        if locked_key not in self._locks:
            raise KeyError("{} was not locked.".format(locked_key))
        self._dir_lockname(locked_key)
        os.rmdir(self._dir_lockname(locked_key))
        self._locks.remove(locked_key)

    @_locked
    def mark_dirty(self, key: str) -> Any:
        """
        Manually mark an entry as dirty. It will be updated on disk on the next occasion.

        :param key: the key that needs to be stored
        :return: the value contained in the key
        """
        val = self[key]
        self[key] = val
        return val

    def _create_lock_file(
        self, path: str, retrycount: int = DB_LOCK_RETRYCOUNT
    ) -> None:
        """
        Create new lock file, waiting up to retrycount seconds.

        :raises: :class:`TimeoutError` if failed.
        :param path: path of the lock file
        :param retrycount: number of tries until raising an Exception
        """
        for i in range(0, retrycount):
            try:
                makedirs(path, exist_ok=False)
                return
            except OSError as ex:
                self.logger.debug(
                    "Waiting for lock on file {} (currently {})".format(path, ex)
                )
                time.sleep(float(2 ** i) / 10)
        raise TimeoutError("Lock for {} could not be acquired in time!".format(path))

    @_locked
    def lock(self, key: str) -> None:
        """
        Wait for a lock.

        :param key: the key to lock
        """
        if key in self._locks:
            raise KeyError("{} already locked".format(key))
        self._create_lock_file(self._dir_lockname(key))
        self._locks.add(key)

    @_locked
    def is_locked(self, key: str) -> bool:
        """
        Return if the key is currently locked by this process.

        :param key: The key
        :return: True if locked by this process, False otherwise
        """
        return key in self._locks

    @_locked
    def reload(self) -> None:
        """
        Reload stored values from disk.

        There is usually no reason to call this.
        Non persisted changes might be lost.
        Only reason would be if another process fiddles with our data concurrently.
        """
        self._cache = {}
        self._dirties = set()
        self._to_delete = set()

    @_locked
    def persist(self) -> None:
        """
        Store all dirty data to disk.

        If no data is to be stored, it's basically free to call.
        """
        for key in self._to_delete:
            locked = self.is_locked(key) or self.ignore_locks
            if not locked:
                self.lock(key)
            os.remove(self._dir_jsonname(key))
            if not locked:
                self.release(key)
            self.logger.debug(f"Deleted {key} from db {self.name}")
        self._to_delete = set()

        for key in self._dirties:
            locked = self.is_locked(key)
            if not locked:
                self.lock(key)
            try:
                with open(self._dir_jsonname(key), "wb") as f:
                    f.write(json.dumps(self._cache[key]).encode("utf-8"))
            finally:
                if not locked:
                    self.release(key)
            self.logger.debug(f"Set {key} in db {self.name}")
        self._dirties = set()

    @_locked
    def __getitem__(self, key: str) -> Any:
        """
        Get an item from the dict. Will hit the cache first, then disk.

        :param key: the key to look up
        :return: the value
        """
        if key in self._to_delete:
            raise KeyError("Key was marked for deletion: {}".format(key))
        try:
            return self._cache[key]
        except KeyError:
            self.logger.debug("Hitting disk for {}".format(key))

        locked = self.is_locked(key) or self.ignore_locks
        if not locked:
            self.lock(key)
        try:
            with open(self._dir_jsonname(key), "rb") as f:
                val = json.loads(f.read().decode("utf-8"))
        except OSError as ex:
            raise KeyError("Key {} not found - {}".format(key, ex))
        finally:
            if not locked:
                self.release(key)

        self._cache[key] = val
        return val

    @_locked
    def __setitem__(self, key: str, value: Any) -> None:
        """
        Set an item. It'll be stored to disk on the next persist.

        :param key: Key to store
        :param value: Value to store
        """
        if key in self._to_delete:
            self._to_delete.remove(key)
        self._cache[key] = value
        self._dirties.add(key)

    @_locked
    def __delitem__(self, key: str) -> None:
        """
        Delete an item. It will be deleted from disk on the next .persist().

        :param key: the key to delete
        """
        self._to_delete.add(key)

    def __iter__(self) -> Iterator[str]:
        """
        Return an iterator over the dict.

        Implicitly persisting the data before reading.

        :return: An iterator containing all keys to a dict.
        """
        self.persist()
        keys = [
            debase64ify(x[len(DB_PREFIX) : -len(DB_EXTENSION)], b"+-")
            for x in os.listdir(self.path)
            if x.startswith(DB_PREFIX) and x.endswith(DB_EXTENSION)
        ]
        yield from keys

    def __len__(self) -> int:
        """
        Calculate the length of the dict.

        Implicitly calls persist.

        :return: the the number of elements
        """
        self.persist()
        keys = [
            x[len(DB_PREFIX) : -len(DB_EXTENSION)]
            for x in os.listdir(self.path)
            if x.startswith(DB_PREFIX) and x.endswith(DB_EXTENSION)
        ]
        return len(keys)



"""Collection of Exception classes to signal the status of the service being checked."""

from abc import ABC
from typing import Optional

from enochecker_core import CheckerTaskResult


class CheckerResult:
    def __init__(
        self, result: CheckerTaskResult, message: Optional[str] = None
    ) -> None:

        if message == "":
            message = None

        self.result = result
        self.message = message

    @staticmethod
    def from_exception(e: Exception) -> "CheckerResult":
        """Converts a given Exception to an extended CheckerResult including Message
        public_message isn't used anywhere yet"""

        if isinstance(e, EnoException):
            message = str(e)
            return CheckerResult(result=e.result, message=message)

        else:
            return CheckerResult(
                CheckerTaskResult.CHECKER_TASK_RESULT_INTERNAL_ERROR, message=None
            )


class EnoException(Exception, ABC):
    """Base error including the Result. Raise a subclass of me once we know what to do."""

    result: CheckerTaskResult = CheckerTaskResult.CHECKER_TASK_RESULT_INTERNAL_ERROR

    def __init__(
        self, message: Optional[str], internal_message: Optional[str] = None,
    ):
        self.message: Optional[str] = message
        self.internal_message: Optional[str] = internal_message

    def __str__(self) -> str:
        return self.message if self.message else ""

    def message_contains(self, flag: Optional[str]) -> bool:
        """" If the string is in the message """
        if not self.message:
            return False
        if not flag:
            return False
        return flag in self.message


class BrokenServiceException(EnoException):
    """Indicates a broken Service."""

    result: CheckerTaskResult = CheckerTaskResult.CHECKER_TASK_RESULT_MUMBLE


class OfflineException(EnoException):
    """Service was not reachable (at least once) during our checks."""

    result: CheckerTaskResult = CheckerTaskResult.CHECKER_TASK_RESULT_OFFLINE


class BrokenCheckerException(EnoException):
    """
    Shouldn't be raised ever since we catch all abstract Errors.

    Used internally if something goes horribly wrong.
    """

    result: CheckerTaskResult = CheckerTaskResult.CHECKER_TASK_RESULT_INTERNAL_ERROR



"""Contains the BaseChecker to be used as base for all checkers."""

import argparse
import datetime
import hashlib
import logging
import os
import socket
import sys
import traceback
import warnings
from abc import ABCMeta, abstractmethod
from concurrent.futures import TimeoutError
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)
from urllib.parse import urlparse

import jsons
from enochecker_cli import add_arguments, task_message_from_namespace
from enochecker_core import CheckerMethod, CheckerTaskMessage, CheckerTaskResult
from flask import Flask

from .checkerservice import init_service
from .logging import ELKFormatter
from .nosqldict import NoSqlDict
from .results import CheckerResult, EnoException, OfflineException
from .storeddict import DB_DEFAULT_DIR, DB_GLOBAL_CACHE_SETTING, StoredDict
from .useragents import random_useragent
from .utils import SimpleSocket, snake_caseify

if TYPE_CHECKING:  # pragma: no cover
    # The import might fail in UWSGI, see the comments below.
    import requests

DEFAULT_TIMEOUT: float = 30
TIME_BUFFER: float = 5  # time in seconds we try to finish earlier

# Global cache for all stored dicts.  TODO: Prune this at some point?
global_db_cache: Dict[str, Union[StoredDict, NoSqlDict]] = {}


def warn_deprecated(old_name: str, new_name: str) -> None:
    """
    Print a warning for the deprecated feature, include the new name in the log
    This needs python development mode or deprecation warnings enabled!
    """
    warnings.warn(
        f"Checker uses deprecated {old_name}; use {new_name} instead.",
        DeprecationWarning,
    )


def parse_args(argv: Optional[Sequence[str]] = None) -> argparse.Namespace:
    """
    Return the parsed argparser args.

    :param argv: argv. Custom argvs. Will default to sys.argv if not provided.
    :return: args object
    """
    if argv is None:
        return parse_args(sys.argv[1:])
    parser = argparse.ArgumentParser(description="Your friendly checker script")

    subparsers = parser.add_subparsers(
        help="The checker runmode (run/listen)", dest="runmode"
    )
    subparsers.required = True

    listen = subparsers.add_parser("listen", help="Spawn checker service")
    listen.add_argument(
        "listen_port", help="The port the checker service should listen on"
    )

    runparser = subparsers.add_parser("run", help="Run checker on cmdline")
    add_arguments(runparser)

    return parser.parse_args(args=argv)


class _CheckerMeta(ABCMeta):
    """
    Some python magic going on right here.

    Each time we subclass BaseChecker, this __init__ is called.
    ABCMeta is used as superclass instead of type, such that BaseChecker is declared abstract -> needs to be overridden.
    """

    def __init__(
        cls: "_CheckerMeta", name: str, bases: Tuple[type, ...], clsdict: Dict[Any, Any]
    ):
        """
        Called whenever this class is subclassed.

        :param name: The name of the new class
        :param bases: Bases classes this class inherits from.
        :param clsdict: Contents of this class (.__dict__)
        """
        if len(cls.mro()) > 2:  # 1==BaseChecker
            cls.service: Flask = init_service(cast(Type[BaseChecker], cls))
        super().__init__(name, bases, clsdict)


class BaseChecker(metaclass=_CheckerMeta):
    """
    All you base are belong to us. Also all your flags. And checker scripts.

    Override the methods given here, then simply init and .run().
    Magic.
    """

    flag_variants: int
    noise_variants: int
    havoc_variants: int

    def __init__(
        self,
        task: CheckerTaskMessage,
        storage_dir: str = DB_DEFAULT_DIR,
        use_db_cache: bool = DB_GLOBAL_CACHE_SETTING,
        json_logging: bool = True,
    ) -> None:
        """
        Init the Checker, fill the params.

        :param task: The CheckerTaskMessage to be executed
        :param storage_dir: The directory to store persistent data in (used by StoredDict)
        :param use_db_cache: whether the DB connections should be cached or a new DB connection should be created per request
        :param json_logging: whether the JSON-based log format for ELK should be used instead of the more human-readable output
        """
        # We import requests after startup global imports may deadlock, see
        # https://github.com/psf/requests/issues/2925
        import requests

        self.requests = requests

        self.time_started_at: datetime.datetime = datetime.datetime.now()
        self.task_id: int = task.task_id
        self.json_logging: bool = json_logging

        self.method: CheckerMethod = task.method
        self.address: str = task.address
        self.team_id: int = task.team_id
        self.team_name: str = task.team_name
        self.current_round_id: int = task.current_round_id
        self.related_round_id: int = task.related_round_id
        self.flag: Optional[str] = task.flag
        self.variant_id: int = task.variant_id
        self.timeout: float = task.timeout / 1000
        self.round_length: float = task.round_length / 1000
        self.task_chain_id: str = task.task_chain_id

        self._noise_cache: Optional[str] = None

        self.storage_dir = storage_dir

        self._setup_logger()
        if use_db_cache:
            self._active_dbs: Dict[str, Union[NoSqlDict, StoredDict]] = global_db_cache
        else:
            self._active_dbs = {}
        self.http_session: requests.Session = self.requests.session()
        self.http_useragent = random_useragent()

        if not hasattr(self, "service_name"):
            self.service_name: str = type(self).__name__.split("Checker")[0]
            self.debug(
                "Assuming checker Name {}. If that's not the case, please override.".format(
                    self.service_name
                )
            )

        if not hasattr(self, "port"):
            self.warning("No default port defined.")
            self.port = -1

        self.debug(
            "Initialized checker for task {} in {} seconds".format(
                jsons.dumps(task), datetime.datetime.now() - self.time_started_at,
            )
        )

    def _setup_logger(self) -> None:
        """
        Set up a logger usable from inside a checker using.

        self.debug, self.info, self.warning, self.error or self.logger
        A logger can have additional args as well as exc_info=ex to log an exception, stack_info=True to log trace.
        """
        self.logger: logging.Logger = logging.Logger(type(self).__name__)
        self.logger.setLevel(logging.DEBUG)

        # default handler
        handler = logging.StreamHandler(sys.stdout)
        handler.setLevel(logging.DEBUG)
        if self.json_logging:
            formatter: logging.Formatter = ELKFormatter(self)
        else:
            formatter = logging.Formatter(
                "%(asctime)s - %(name)s - %(levelname)s - %(message)s"
            )
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

        self.debug: Callable[..., None] = self.logger.debug
        self.info: Callable[..., None] = self.logger.info
        self.warning: Callable[..., None] = self.logger.warning
        self.error: Callable[..., None] = self.logger.error
        self.critical: Callable[..., None] = self.logger.critical

    @property
    def chain_db(self) -> Any:
        """
        get the team_db entry for the current chain. Short hand version for self.team_db[self.task_chain_id]

        :return: the team_db entry
        """
        return self.team_db[self.task_chain_id]

    @chain_db.setter
    def chain_db(self, value: Any) -> None:
        """
        set the team_db entry for the current task chain. Short hand version for self.team_db[self.task_chain_id] = value

        :param value: the value to set in the team_db
        """
        self.team_db[self.task_chain_id] = value

    @property
    def noise(self) -> str:
        """
        Creates a stable noise value for the current task chain.
        Do not use for indexing (use self.task_chain_id instead).

        :return: A noise string, unique for each task_chain_id.
        """
        if self._noise_cache is None:
            if not self.task_chain_id:
                self.warning("No valid task_chain_id when calling noise!")
                return "<none>"
            # We cache the hex in case it's called often.
            m = hashlib.sha256()
            m.update(self.task_chain_id.encode())
            self._noise_cache = m.hexdigest()
        return self._noise_cache

    @property
    def time_running(self) -> float:
        """
        How long this checker has been running for.

        :return: time this checker has been running for in seconds
        """
        return (datetime.datetime.now() - self.time_started_at).total_seconds()

    @property
    def time_remaining(self) -> float:
        """
        Return a remaining time that is safe to be used as timeout.

        Includes a buffer of TIME_BUFFER seconds.

        :return: A safe number of seconds that may still be used
        """
        return max(
            getattr(self, "timeout", DEFAULT_TIMEOUT) - self.time_running - TIME_BUFFER,
            1,
        )

    # ---- Basic checker functionality ---- #

    def _run_method(self, method: CheckerMethod) -> Optional[CheckerTaskResult]:
        """
        Execute a checker method, pass all exceptions along to the calling function.

        :param method: When calling run, you may call a different method than the one passed on Checker creation
                        using this optional param.
        :return: a CheckerTaskResult enum value
        """

        return getattr(self, snake_caseify(method.value))()

    def run(self, method: Optional[CheckerMethod] = None) -> CheckerResult:
        """
        Execute the checker and catch errors along the way.

        :param method: When calling run, you may call a different method than the one passed on Checker creation
                        using this optional param.
        :return: A CheckerResult as a representation of the CheckerResult response as definded in the Spec.
        """
        if method is None:
            method = self.method

        try:
            ret = self._run_method(method)
            if ret is not None:
                warnings.warn(
                    "Returning a result is not recommended and will be removed in the future. Raise EnoExceptions with additional text instead.",
                    DeprecationWarning,
                )
                try:
                    CheckerTaskResult(ret)
                except:
                    self.error(
                        "Illegal return value from {}: {}".format(self.method, ret)
                    )
                    return CheckerResult(
                        CheckerTaskResult.CHECKER_TASK_RESULT_INTERNAL_ERROR
                    )

                ret = CheckerTaskResult(ret)
                self.info("Checker [{}] resulted in {}".format(self.method, ret.name))
                return CheckerResult(ret)

            # Returned Normally
            self.info("Checker [{}] executed successfully!".format(self.method))
            return CheckerResult(CheckerTaskResult.CHECKER_TASK_RESULT_OK)

        except EnoException as eno:
            stacktrace = "".join(
                traceback.format_exception(None, eno, eno.__traceback__)
            )
            self.info(
                "Checker[{}] result: {}({})".format(
                    self.method, eno.result.name, stacktrace
                ),
                exc_info=eno,
            )

            if eno.message_contains(self.flag):
                self.error(f"EnoMessage contained flag! (Exception was {eno}")
                eno.message = None
            if eno.internal_message:
                self.info(f"Internal info for return: {eno.internal_message}")

            return CheckerResult.from_exception(eno)
        except self.requests.HTTPError as ex:
            self.info("Service returned HTTP Errorcode [{}].".format(ex), exc_info=ex)
            return CheckerResult(
                CheckerTaskResult.CHECKER_TASK_RESULT_MUMBLE,
                "Service returned HTTP Error",
            )
        except EOFError as ex:
            self.info("Service returned EOF error [{}].".format(ex), exc_info=ex)
            return CheckerResult(
                CheckerTaskResult.CHECKER_TASK_RESULT_MUMBLE,
                "Service returned EOF while reading response.",
            )
        except (
            self.requests.ConnectionError,  # requests
            self.requests.exceptions.ConnectTimeout,  # requests
            TimeoutError,
            socket.timeout,
            ConnectionError,
            OSError,
            ConnectionAbortedError,
        ) as ex:
            self.info(
                "Error in connection to service occurred: {}\n".format(ex), exc_info=ex
            )
            return CheckerResult(
                CheckerTaskResult.CHECKER_TASK_RESULT_OFFLINE,
                message="Error in connection to service occured",
            )  # , ex.message
        except Exception as ex:
            stacktrace = "".join(traceback.format_exception(None, ex, ex.__traceback__))
            self.error(
                "Unhandled checker error occurred: {}\n".format(stacktrace), exc_info=ex
            )
            return CheckerResult.from_exception(ex)  # , ex.message
        finally:
            for db in self._active_dbs.values():
                # A bit of cleanup :)
                db.persist()

    @abstractmethod
    def putflag(self) -> None:
        """
        Store a flag in the service.

        In case multiple flags are provided, self.variant_id gives the appropriate flag store to target.
        The flag itself can be retrieved from self.flag.
        On error, raise an Eno Exception.

        :raises: EnoException on error
        """
        pass

    @abstractmethod
    def getflag(self) -> None:
        """
        Retrieve a flag from the service.

        Use self.flag to get the flag that needs to be recovered and self.task_chain_id as a key to retrieve data from the database stored during putflag
        On error, raise an EnoException.

        :raises: EnoException on error
        """
        pass

    @abstractmethod
    def putnoise(self) -> None:
        """
        Store noise in the service.

        The noise should later be recoverable.
        The difference between noise and flag is that noise does not have to remain secret for other teams.
        This method can be called multiple times per round. Check which variant is called using self.variant_id.
        On error, raise an EnoException.

        :raises: EnoException on error
        """
        pass

    @abstractmethod
    def getnoise(self) -> None:
        """
        Retrieve noise in the service.

        The noise to be retrieved, can be restored from the database by using the self.task_chain_id used to store it during putflag.
        The difference between noise and flag is, tht noise does not have to remain secret for other teams.
        This method can be called many times per round. Check which variant is called using self.variant_id.
        On error, raise an EnoException.

        :raises: EnoException on error
        """
        pass

    @abstractmethod
    def havoc(self) -> None:
        """
        Unleash havoc on the app -> Do whatever you must to prove the service still works. Or not.

        On error, raise an EnoException.

        :raises: EnoException on Error
        """
        pass

    @abstractmethod
    def exploit(self) -> None:
        """
        Use this method strictly for testing purposes.

        Will hopefully not be called during the actual CTF.

        :raises: EnoException on Error
        """
        pass

    # ---- DB specific methods ---- #
    def db(
        self, name: str, ignore_locks: bool = False
    ) -> Union[NoSqlDict, StoredDict]:  # TODO: use a common supertype for all backends
        """
        Get a (global) db by name.

        Subsequent calls will return the same db.
        Names can be anything, for example the team name, round numbers etc.

        :param name: The name of the DB
        :param ignore_locks: Should only be set if you're sure-ish keys are never shared between instances.
                Manual locking ist still possible.
        :return: A dict that will be self storing. Alternatively,
        """
        try:
            db = self._active_dbs[name]
            db.logger = self.logger
            # TODO: Settng a new Logger backend may throw logs in the wrong direction in a multithreaded environment!
            return db
        except KeyError:
            checker_name = type(self).__name__
            self.debug("Remote DB {} was not cached.".format(name))
            if os.getenv("MONGO_ENABLED"):
                host = os.getenv("MONGO_HOST")
                port = os.getenv("MONGO_PORT")
                username = os.getenv("MONGO_USER")
                password = os.getenv("MONGO_PASSWORD")
                self.debug(
                    "Using NoSqlDict mongo://{}:{}@{}:{}".format(
                        username,
                        "".join(["X" for c in password]) if password else "None",
                        host,
                        port,
                    )
                )

                ret: Union[NoSqlDict, StoredDict] = NoSqlDict(
                    name=name,
                    checker_name=checker_name,
                    host=host,
                    port=port,
                    username=username,
                    password=password,
                    logger=self.logger,
                )
            else:
                self.debug(
                    "MONGO_ENABLED not set. Using stored dict at {}".format(
                        self.storage_dir
                    )
                )
                ret = StoredDict(
                    name=name,
                    base_path=self.storage_dir,
                    ignore_locks=ignore_locks,
                    logger=self.logger,
                )
            self._active_dbs[name] = ret
            return ret

    @property
    def global_db(
        self,
    ) -> Union[NoSqlDict, StoredDict]:  # TODO: use a common supertype for all backends
        """
        Get a global storage shared between all teams and rounds.

        Subsequent calls will return the same db.
        Prefer db_team_local or db_round_local

        :return: The global db
        """
        return self.db("global")

    @property
    def team_db(
        self,
    ) -> Union[NoSqlDict, StoredDict]:  # TODO: use a common supertype for all backends
        """
        Return the database for the current team.

        :return: The team local db
        """
        return self.get_team_db()

    def get_team_db(
        self, team_id: Optional[int] = None
    ) -> Union[NoSqlDict, StoredDict]:  # TODO: use a common supertype for all backends
        """
        Return the database for a specific team.

        Subsequent calls will return the same db.

        :param team: Return a db for an other team. If none, the db for the local team will be returned.
        :return: The team local db
        """
        team = team_id if team_id is not None else self.team_id
        return self.db("team_{}".format(team), ignore_locks=True)

    # ---- Networking specific methods ---- #
    def _sanitize_url(
        self, route: str, port: Optional[int] = None, scheme: Optional[str] = None
    ) -> str:
        if port is None:
            port = self.port
        if port is None:
            raise ValueError("Port for service not set. Cannot Request.")

        if ":" in self.address:
            netloc = "[{}]:{}".format(self.address, port)
        else:
            netloc = "{}:{}".format(self.address, port)
        if scheme is None:
            url = urlparse(route)
        else:
            url = urlparse(route, scheme=scheme)
        # noinspection PyProtectedMember
        return url._replace(netloc=netloc).geturl()

    def connect(
        self,
        host: Optional[str] = None,
        port: Optional[int] = None,
        timeout: Optional[float] = None,
        retries: int = 3,
    ) -> SimpleSocket:
        """
        Open a socket/telnet connection to the remote host.

        Use connect(..).get_socket() for the raw socket.

        :param host: the host to connect to (defaults to self.address)
        :param port: the port to connect to (defaults to self.port)
        :param timeout: timeout on connection (defaults to self.timeout)
        :param retries: the amount of times this socket connection should be retried
        :return: A connected Telnet instance
        """

        if timeout is not None:

            def timeout_fun() -> float:
                return cast(float, timeout)

        else:

            def timeout_fun() -> float:
                return self.time_remaining / 2

        if port is None:
            port = self.port
        if host is None:
            host = self.address

        if retries < 0:
            raise ValueError("Number of retries must be greater than zero.")

        for i in range(0, retries + 1):  # + 1 for the initial try
            try:

                timeout = timeout_fun()
                self.debug(
                    "Opening socket to {}:{} (timeout {} secs).".format(
                        host, port, timeout
                    )
                )
                return SimpleSocket(
                    host,
                    port,
                    timeout=timeout,
                    logger=self.logger,
                    timeout_fun=timeout_fun,
                )

            except Exception as e:
                self.warning(
                    f"Failed to establish connection to {host}:{port}, Try #{i+1} of {retries+1} ",
                    exc_info=e,
                )
                continue

        self.error(f"Failed to establish connection to {host}:{port}")
        raise OfflineException("Failed establishing connection to service.")

    @property
    def http_useragent(self) -> str:
        """
        Return the useragent for http(s) requests.

        :return: the current useragent
        """
        return self.http_session.headers["User-Agent"]

    @http_useragent.setter
    def http_useragent(self, useragent: str) -> None:
        """
        Set the useragent for http requests.

        Randomize using http_useragent_randomize()

        :param useragent: the useragent
        """
        self.http_session.headers["User-Agent"] = useragent

    def http_useragent_randomize(self) -> str:
        """
        Choose a new random http useragent.

        Note that http requests will be initialized with a random user agent already.
        To retrieve a random useragent without setting it, use random instead.

        :return: the new useragent
        """
        new_agent = random_useragent()
        self.http_useragent = new_agent
        return new_agent

    def http_post(
        self,
        route: str = "/",
        params: Any = None,
        port: Optional[int] = None,
        scheme: str = "http",
        raise_http_errors: bool = False,
        timeout: Optional[float] = None,
        **kwargs: Any,
    ) -> "requests.Response":
        """
        Perform a (http) requests.post to the current host.

        Caches cookies in self.http_session

        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        return self.http(
            "post", route, params, port, scheme, raise_http_errors, timeout, **kwargs
        )

    def http_get(
        self,
        route: str = "/",
        params: Any = None,
        port: Optional[int] = None,
        scheme: str = "http",
        raise_http_errors: bool = False,
        timeout: Optional[float] = None,
        **kwargs: Any,
    ) -> "requests.Response":
        """
        Perform a (http) requests.get to the current host.

        Caches cookies in self.http_session

        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        return self.http(
            "get", route, params, port, scheme, raise_http_errors, timeout, **kwargs
        )

    def http(
        self,
        method: str,
        route: str = "/",
        params: Any = None,
        port: Optional[int] = None,
        scheme: str = "http",
        raise_http_errors: bool = False,
        timeout: Optional[float] = None,
        **kwargs: Any,
    ) -> "requests.Response":
        """
        Perform an http request (requests lib) to the current host.

        Caches cookies in self.http_session

        :param method: The request method
        :param params: The parameter
        :param route: The route
        :param port: The remote port in case it has not been specified at creation
        :param scheme: The scheme (defaults to http)
        :param raise_http_errors: If True, will raise exception on http error codes (4xx, 5xx)
        :param timeout: How long we'll try to connect (default: self.timeout)
        :return: The response
        """
        kwargs.setdefault("allow_redirects", False)
        url = self._sanitize_url(route, port, scheme)
        if timeout is None:
            timeout = self.time_remaining / 2
        self.debug(
            "Request: {} {} with params: {} and {} secs timeout.".format(
                method, url, params, timeout
            )
        )
        resp = self.http_session.request(
            method, url, params=params, timeout=timeout, **kwargs
        )
        if raise_http_errors:
            resp.raise_for_status()
        return resp


def run(
    checker_cls: Type[BaseChecker], args: Optional[Sequence[str]] = None,
) -> Optional[CheckerResult]:
    """
    Run a checker, either from cmdline or as uwsgi script.

    :param checker: The checker (subclass of basechecker) to run
    :param force_service: if True (non-default), the server will skip arg parsing and immediately spawn the web service.
    :param args: optional parameter, providing parameters
    :return:  Never returns.
    """
    parsed = parse_args(args)
    if parsed.runmode == "listen":
        checker_cls.service.run(host="::", debug=True, port=parsed.listen_port)
        return None
    else:
        task_message = task_message_from_namespace(parsed)
        result = checker_cls(task_message).run()
        print(f"Checker run resulted in Result: {result.result.name}")
        return result



"""Backend for team_db based on MongoDB."""

import logging
from collections.abc import MutableMapping
from functools import wraps
from threading import RLock, current_thread
from typing import TYPE_CHECKING, Any, Callable, Dict, Iterable, Optional, Union

from . import utils
from .utils import base64ify

if TYPE_CHECKING:  # pragma: no cover
    # We do these things as late as possible to avoid strange deadlocks when forking in UWSGI
    from pymongo import MongoClient

# LOGGING SETUP
logging.basicConfig(level=logging.DEBUG)
dictlogger = logging.Logger(__name__)
dictlogger.setLevel(logging.DEBUG)

RETRY_COUNT = 4

# DB DEFAULT PARAMS
DB_DEFAULT_USER = None
DB_DEFAULT_PASS = None
DB_DEFAULT_HOST = "localhost"
DB_DEFAULT_PORT = 27017


def to_keyfmt(key: Any) -> str:
    """
    Convert a string to a key used in the MongoDB.

    Currently this only returns the string representation of the key.

    :param key: the key to format
    :return: string representation of the key
    """
    return str(key)  # + type(key).__name__


def _try_n_times(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def try_n_times(self: "NoSqlDict", *args: Any, **kwargs: Any) -> Any:
        from pymongo.errors import PyMongoError

        for i in range(RETRY_COUNT):
            try:
                return func(self, *args, **kwargs)
            except PyMongoError as ex:
                self.logger.error("noSQLdict_Error, Try {}".format(str(i)), exc_info=ex)
                if i == RETRY_COUNT:
                    raise
        assert False  # this should never happen

    return try_n_times


class NoSqlDict(MutableMapping):
    """A dictionary that is MongoDb backed."""

    dblock = RLock()

    @classmethod
    def get_client(
        cls,
        host: str,
        port: int,
        username: Optional[str],
        password: Optional[str],
        logger: logging.Logger,
    ) -> "MongoClient":
        """
        Lazily try to get the mongo db connection or creates a new one.

        :param host: mongo host
        :param port: mongo port
        :param username: the username to connect to
        :param password: the password to use
        :return:
        """
        mongo_name = utils.ensure_valid_filename(
            "mongo_{}_{}_{}_{}".format(host, port, username, password)
        )
        if hasattr(cls, mongo_name):
            return getattr(cls, mongo_name)

        from pymongo import MongoClient

        with NoSqlDict.dblock:
            if hasattr(cls, mongo_name):
                # we found a mongo in the meantime
                return getattr(cls, mongo_name)
            mongo = MongoClient(
                host=host, port=port, username=username, password=password
            )
            setattr(cls, mongo_name, mongo)
        logger.debug(
            "MONGO CLIENT INITIALIZED for thread {}: {}".format(current_thread(), mongo)
        )
        return mongo

    def __init__(
        self,
        name: str = "default",
        checker_name: str = "BaseChecker",
        host: Optional[str] = None,
        port: Union[int, str, None] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        logger: Optional[logging.Logger] = None,
        *args: Any,
        **kwargs: Any,
    ):
        """
        Initialize a NoSqlDict with specified MongoDB backend.

        :param name: name of the backend
        :param checker_name: name of the checker
        :param host: MongoDB host
        :param port: MongoDB port
        :param username: MongoDB username
        :param password: MongoDB password
        """
        if logger:
            self.logger = logger
        else:
            self.logger = dictlogger

        self.dict_name = base64ify(name, altchars=b"-_")
        self.checker_name = checker_name
        self.cache: Dict[Any, Any] = {}
        host_: str = host or DB_DEFAULT_HOST
        if isinstance(port, int):
            port_: int = port
        else:
            port_ = int(port or DB_DEFAULT_PORT)
        username_: Optional[str] = username or DB_DEFAULT_USER
        password_: Optional[str] = password or DB_DEFAULT_PASS
        self.db = self.get_client(host_, port_, username_, password_, self.logger)[
            checker_name
        ][self.dict_name]
        try:
            self.db.index_information()["checker_key"]
        except KeyError:
            self.db.create_index(
                [("key", 1)], name="checker_key", unique=True, background=True
            )

    @_try_n_times
    def __setitem__(self, key: Any, value: Any) -> None:
        """
        Set an entry in the dictionary.

        Caches the value for future gets and stores it in the MongoDB.

        :param key: key in the dictionary
        :param value: value in the dictionary
        """
        self.cache[key] = value

        query_dict = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }

        to_insert = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
            "value": value,
        }

        self.db.replace_one(query_dict, to_insert, upsert=True)

    @_try_n_times
    def __getitem__(self, key: Any, print_result: bool = False) -> Any:
        """
        Get an entry from the dictionary.

        Returns values from cache when they were inserted in the same checker execution.

        :param key: key of the value to retrieve
        :param print_result: TODO
        :return: retrieved value
        """
        if key in self.cache.items():
            return self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }

        result = self.db.find_one(to_extract)

        if print_result:
            self.logger.debug(result)

        if result:
            self.cache[key] = result["value"]
            return result["value"]
        raise KeyError("Could not find {} in {}".format(key, self))

    @_try_n_times
    def __delitem__(self, key: Any) -> None:
        """
        Delete an entry from the dictionary.

        Also deletes the value from the cache in addition to the MongoDB backend.

        :param key: key to delete
        """
        if key in self.cache:
            del self.cache[key]

        to_extract = {
            "key": to_keyfmt(key),
            "checker": self.checker_name,
            "name": self.dict_name,
        }
        self.db.delete_one(to_extract)

    @_try_n_times
    def __len__(self) -> int:
        """
        Return the number of elements in the dictionary.

        :return: number of elements
        """
        return self.db.count_documents(
            {"checker": self.checker_name, "name": self.dict_name}
        )

    @_try_n_times
    def __iter__(self) -> Iterable[Any]:
        """
        Return an iterator over all entries in the dictionary.

        :return: iterator over the entries
        """
        iterdict = {"checker": self.checker_name, "name": self.dict_name}
        results = self.db.find(iterdict)
        yield from map(lambda res: res["key"], results)

    def persist(self) -> None:
        """
        Persist the changes in the backend.

        Currently this function does nothing since every access queries the database already.
        In the future changes to the DB might be cached until they are explicitly persisted.
        """
        # TODO: could wait until here before hitting the mongodb...
        pass



from .enochecker import BaseChecker, parse_args, run  # the BaseChecker
from .results import (  # Possible results
    BrokenServiceException,
    EnoException,
    OfflineException,
)
from .utils import (  # the util stuff
    SimpleSocket,
    assert_equals,
    assert_in,
    base64ify,
    debase64ify,
    ensure_bytes,
    ensure_unicode,
    ensure_valid_filename,
    serve_once,
    sha256ify,
    snake_caseify,
    start_daemon,
)

name = "enochecker"
